package CBuffer(CBuffer(..), CBToken, mkCBuffer) where

-- A "Completion Buffer", parameterized by
--    n, the number of items
--    a, the type of an item
-- A completion buffer is like a FIFO, except that items
-- can be entered out-of-order.
-- getToken  reserves a spot in the FIFO.
-- done      enters an item in the FIFO in the reserved spot.
-- get       retrieves the next item in order.

import Array
import Counter

interface CBuffer n a =
    getToken    :: CBToken n
    getTokenAck :: Action
    done        :: CBToken n -> a -> Action
    get         :: a
    ack         :: Action

data CBToken n = CBToken (Bit (TLog n))
    deriving (Bits)

mkCBuffer :: (Bits a sa, Log n ln, Log n (TLog n), Add 1 ln ln1) =>
             Module (CBuffer n a)
mkCBuffer =
  module
    let hi = fromInteger (valueOf n - 1)
	incr r =
	    r := if r == hi then 0 else r + 1
    buf :: Array (Bit ln) (Maybe a) <- mkArrayWCF 0 hi
    i :: Reg (Bit ln) <- mkReg 0	-- input index
    o :: Reg (Bit ln) <- mkReg 0	-- output index
    n :: Counter ln1  <- mkCounter 0	-- number of filled slots
    interface CBuffer
	    getToken = CBToken i
	      when n.value <= hi
	    getTokenAck =
		action
		    incr i
		    n.up
		    buf.upd i Nothing
	      when n.value <= hi
	    done (CBToken t) a = buf.upd t (Just a)
	    get = x
	      when n.value > 0, Just x <- buf.sub o
	    ack =
		action
		    incr o
		    n.down
	      when n.value > 0, Just x <- buf.sub o
