package LPM(mkLPM) where

-- A "Longest Prefix Match" module.
-- Implemented as a straight pipeline that performs from
-- one to three memory accesses per IP address, using
-- 16bits, 8bits and 8bits respectively to navigate
-- the sparse tree.

import FIFO
import ROM

import ROMports

import LPMifc

data Mid = Lookup IPaddr | Done LuResult
    deriving (Bits)

mkLPM :: AsyncROM lat LuAddr LuData -> Module LPM
mkLPM rom =
    module
	(rom0, rom1, rom2) <- mk3ROMports rom

	fifo0 :: FIFO Mid <- mkFIFO
	fifo1 :: FIFO Mid <- mkFIFO
	fifo2 :: FIFO Mid <- mkFIFO
	ofifo :: FIFO LuResult <- mkFIFO

	rules
	  "Stage 1: lookup, leaf":
	    when Lookup ipa <- fifo0.first, Leaf res <- rom0.result
	     ==> action
		    fifo0.deq
		    rom0.ack
		    fifo1.enq (Done res)

	  "Stage 1: lookup, node":
	    when Lookup ipa <- fifo0.first, Node addr <- rom0.result
	     ==> action
		    fifo0.deq
		    rom0.ack
		    rom1.read (addr + zeroExtend ipa[31:24])
		    fifo1.enq (Lookup (ipa << 8))

	  "Stage 2: no op":
	    when (Done res) <- fifo1.first
	     ==> action
		    fifo1.deq
		    fifo2.enq (Done res)
	  "Stage 2: lookup, leaf":
	    when (Lookup ipa) <- fifo1.first, Leaf res <- rom1.result
	     ==> action
		    fifo1.deq
		    rom1.ack
		    fifo2.enq (Done res)
	  "Stage 2: lookup, node":
	    when (Lookup ipa) <- fifo1.first, Node addr <- rom1.result
	     ==> action
		    fifo1.deq
		    rom1.ack
		    rom2.read (addr + zeroExtend ipa[31:24])
		    fifo2.enq (Lookup (ipa << 8))

	  "Completion: no op":
	    when (Done res) <- fifo2.first
	     ==> action
		    fifo2.deq
		    ofifo.enq res
	  "Completion: lookup, leaf":
	    when (Lookup _) <- fifo2.first, Leaf res <- rom2.result
	     ==> action
		    fifo2.deq
		    rom2.ack
		    ofifo.enq res
	  "Completion: lookup":
	    when (Lookup _) <- fifo2.first, Node addr <- rom2.result
	     ==> action
		    fifo2.deq
		    rom2.ack
		    ofifo.enq badTable		-- should never happen

	interface
	    -- Stage 0
	    luReq ipa =
		action
		    rom0.read (zeroExtend ipa[31:16])
		    fifo0.enq (Lookup (ipa << 16))
	    luResp = ofifo.first
	    luRespAck = ofifo.deq
