package LPMK(mkLPMK) where

-- A "Longest Prefix Match" module.
-- Implemented as a circular pipeline that performs from
-- one to three memory accesses per IP address, using
-- 16bits, 8bits and 8bits respectively to navigate
-- the sparse tree.

import FIFO
import CBuffer

import ROM
import ROMports

import LPMifc

type NStage = 7

mkLPMK :: AsyncROM lat LuAddr LuData -> Module LPM
mkLPMK rom =
    module
	cb :: CBuffer NStage LuResult <- mkCBuffer
	tf :: FIFO (CBToken NStage, IPaddr) <- mkFIFO
	ops :: FIFO (CBToken NStage, IPaddr) <- mkSizedFIFO (valueOf lat + 1)
	buf :: FIFO ((CBToken NStage, IPaddr), LuAddr) <- mkSizedFIFO (valueOf NStage + 1)
	rules
	  "Complete":
	    when (Leaf res) <- rom.result, (tok, ipa) <- ops.first
	     ==> action
		    rom.ack
		    ops.deq
		    cb.done tok res
	  "Recirc":
	    when (Node addr) <- rom.result, (tok, ipa) <- ops.first
	     ==> action
		    rom.ack
		    ops.deq
		    buf.enq ((tok, ipa << 8), addr + zeroExtend ipa[31:24])
	  "Move":
	    when (tokipa, addr) <- buf.first
	     ==> action
		    buf.deq
		    rom.read addr
		    ops.enq tokipa
	  -- XXX Enter should have lowest priority
	  "Enter":
	    when (tok, ipa) <- tf.first
	     ==> action
		    tf.deq
		    rom.read (zeroExtend ipa[31:16])
		    ops.enq (tok, ipa << 16)
	interface
	    luReq ipa =
		action
		    tf.enq (cb.getToken, ipa)
		    cb.getTokenAck
	    luResp = cb.get
	    luRespAck = cb.ack
