package ROMports(mk3ROMports) where

-- A "port replicator" that converts a single ROM port
-- into 3 ROM ports.
-- The port(s) are parameterized by latency, the type of
-- the address, and the type of the resulting data.
-- Note: it is possible in Bluespec to write a more general
-- version that is also parameterized by the number of ports.

import FIFO
import Counter

import ROM

type Tag = Bit 2	-- Enough for 3 ports

mk3ROMports :: (Bits dta dtas, Add lat 1 lat1, Log lat1 llat) =>
	       AsyncROM lat adr dta -> Module (AsyncROM lat1 adr dta,
                                               AsyncROM lat1 adr dta,
                                               AsyncROM lat1 adr dta)
mk3ROMports rom =
  module
    let ilat = fromInteger (valueOf lat)
    tags :: FIFO Tag <- mkSizedFIFO ilat
    let mkPort :: Tag -> Module (AsyncROM lat1 adr dta)
	mkPort i =
	  module
	    out :: FIFO dta <- mkSizedFIFO ilat
	    cnt :: Counter llat <- mkCounter ilat
	    rules
		when tags.first == i
		 ==> action
		        tags.deq
		        rom.ack
		        out.enq rom.result
	    interface
		read a = action
			    rom.read a
			    tags.enq i
			    cnt.down
		    when cnt.value > 0
		result = out.first
		ack = action
			    out.deq
			    cnt.up

    port0 <- mkPort 0
    port1 <- mkPort 1
    port2 <- mkPort 2
    interface (port0, port1, port2)
